// Copyright (C) 2001 Eric Hanchrow
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eric Hanchrow -- offby1@blarg.net
// 1914 N. 40th
// Seatlle, WA, USA
// 98103-8310

#include <vector>
#include "auction.hh"
#include "call.hh"
#include "compass.hh"
#include "exceptions.hh"

namespace {
  // This code smells.  I suspect logic similar to this lurks just
  // below, in the get_contract method.
  template<class C> seat last_caller (const seat &dealer, const std::vector<const call *> &calls)
  {
    for (std::vector<const call *>::const_iterator i = calls.end () - 1;
         i >= calls.begin ();
         i--)
      {
        if (dynamic_cast<const C *>(*i))
          return dealer + (i - calls.begin ());
      }

    Assert<internal_error>(false, "No calls");
  }
};

void auction::push_back (const call *c)
{
  Assert<Bad_State> (in_progress == _state);

  if (const bid *bp = dynamic_cast<const bid*>(c))
    {
      if (_highest_bid
          &&
          !(*_highest_bid < *bp))

        throw Insufficient_Bid ();

      _f_double_active = false;

      if (_highest_bid)
        delete _highest_bid;

      _highest_bid = new bid (*bp);
    }

  // If this call was a pass, the auction might be over.
  else if (dynamic_cast<const pass *>(c))
    {
      // If there were exactly three calls already, and they're all
      // passes, then this hand has been passed out.
      if ((_calls.size () == 3)
          && (dynamic_cast<const pass *>(_calls[0]))
          && (dynamic_cast<const pass *>(_calls[1]))
          && (dynamic_cast<const pass *>(_calls[2])))
        {
          _state = passed_out;
        }
      // Otherwise, if the last two calls were passes, then the
      // contract is settled.
      else if ((_calls.size () >= 3)
               &&
               (dynamic_cast<const pass *>(_calls[_calls.size () - 1]))
               &&
               (dynamic_cast<const pass *>(_calls[_calls.size () - 2])))
        {
          // Now would be a good time to determine the contract, and
          // save it in a private member; that way, get_contract
          // wouldn't need to assert that _state == contract_settled.
          _state = contract_settled;
        }
    }
  else
    {
      const Double *pD = dynamic_cast<const Double *>(c);
      Assert<internal_error>(pD, "A call wasn't a pass, bid, or double");

      _f_redouble = (4 == pD->multiplier ());

      // Can't double an active double, but you can *re*double an
      // active double.
      Assert<Bad_Double>(_f_double_active == _f_redouble);

      // Can't double or redouble unless somebody has bid.
      Assert<Bad_Double>(_highest_bid);

      const seat current_caller = dealer + _calls.size ();

      // Can't double your partner, but you can *re*double your partner.
      Assert<Bad_Double>(_f_redouble
                         ==
                         current_caller.on_same_team (last_caller<bid> (dealer, _calls)));

      _f_double_active = true;
      _doubler = current_caller;
    }

  _calls.push_back (c);

}

contract auction::get_contract () const
{
  Assert<Bad_State> (this->_state == contract_settled);

  bid        rv_bid;

  // I wish I could think of a shorter name that conveys all this
  // information.
  seat       guy_on_declarers_team_who_bid_final_suit;
  bool       f_found_declarer (false); // debugging only

  // Search all the calls, starting at the end and working backward,
  // until we find an actual bid.  Figure out which seat made that
  // bid.  See if they're vulnerable.
  bool f_found_a_bid = false;
  std::vector<const call *>::const_iterator i;
  for (i = _calls.end () - 1;
       i >= _calls.begin ();
       i--)
    {
      if (const bid *final_bid = dynamic_cast<const bid *>(*i))
        {
          f_found_a_bid = true;

          rv_bid = *final_bid;

          // Now we have to see which of the two partners first
          // mentioned that suit.
          for (std::vector<const call *>::const_iterator j = i;
               j >= _calls.begin ();
               j--)
            {
              if ((i - j) % 2)
                // This call was made by an opponent of the declarer,
                // so we ignore it.
                continue;

              const bid *bp = dynamic_cast<const bid *>(*j);
              if (bp
                  &&
                  bp->denom () == final_bid->denom ())
                {
                  // This assignment might happen more than once,
                  // which is fine; we care only about the final
                  // assignment.  That value will be the *first*
                  // partner on declarer's team who bid the final
                  // suit.  First, because we're traversing the
                  // auction backwards, and it's the *last* item we
                  // saw going backwards.
                  guy_on_declarers_team_who_bid_final_suit = this->dealer + (j - _calls.begin ());
                  f_found_declarer = true;
                }
            }

          Assert<No_Declarer> (f_found_declarer);

          break;
        }
    }

  Assert<No_Bids> (f_found_a_bid);

  contract::multiplier rv_multiplier = contract::not_doubled;

  if (_f_double_active)
    {
      rv_multiplier = (_f_redouble
                       ? contract::redoubled
                       : contract::doubled);
    }

  return contract (rv_bid,
                   guy_on_declarers_team_who_bid_final_suit,
                   rv_multiplier);
}

auction::~auction ()
{
  for (std::vector<const call *>::iterator i = _calls.begin ();
       i != _calls.end ();
       i++)
    {
      if (*i)
        {
          delete *i;
          *i = 0;
        }
    }

  if (_highest_bid)
    delete _highest_bid;
}
