// Copyright (C) 2001 Eric Hanchrow
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eric Hanchrow -- offby1@blarg.net
// 1914 N. 40th
// Seatlle, WA, USA
// 98103-8310
#ifndef FILE_CALL_SEEN
#define FILE_CALL_SEEN

#include <cassert>
#include <iostream>
#include <string>

#include "card.hh"

class call
{
public:
  virtual void print () const = 0;
  virtual ~call () {}
  friend std::ostream & operator<<(std::ostream &o, const call &c);
};

class pass: public call
{
public:
  void print () const { std::cout << "pass"; }
};

class denomination
{
  suit _s;
  bool _f_notrump;
public:
  denomination (const suit &s) : _s (s), _f_notrump (false) {}
  denomination () : _f_notrump (true) {}
  denomination & operator=(const suit &s) { _s = s; _f_notrump = false; return *this; }

  void make_notrump() { _f_notrump = true; }

  bool is_notrump () const { return _f_notrump; }

  /// Every call to get_suit ought to be preceded by a call to
  /// is_notrump!  Or else you risk hitting this assertion.
  suit get_suit   () const { assert (!_f_notrump); return _s; }

  bool operator==(const denomination &other) const
  {
    // Make sure we don't examine _s if _f_notrump is set -- by
    // definition, _s is garbage in that case.
    if (this->_f_notrump && other._f_notrump)
      return true;

    return ((this->_f_notrump == other._f_notrump)
            &&
            (this->_s == other._s));
  }

  bool operator<(const denomination &other) const
  {
    if (this->_f_notrump != other._f_notrump)
      return other._f_notrump;
    else if (this->_f_notrump)
      return false;
    else
      return (this->_s < other._s);
  }
  friend std::ostream & operator<< (std::ostream &o, const denomination &d);
};


class bid: public call
{
  unsigned short _level;
  denomination _denomination;
public:
  bid (unsigned short l, const denomination &d) : _level (l), _denomination (d)
  {
    assert ((0 < _level)
            &&
            (_level < 8));
  }

  bid () : _level (1), _denomination (clubs)
  {
    assert ((0 < _level)
            &&
            (_level < 8));
  }

  unsigned short level () const { return _level;        }
  denomination   denom () const { return _denomination; }
  void print () const { std::cout << _level << " " << _denomination; }

  /**
   * This is (probably) only used by the ultra-boneheaded bidding
   * algorithm
   */

  bid next () const
  {
    if ((7 == _level)
        &&
        _denomination.is_notrump ())
      {
        throw "There *is* no bid higher than seven notrump";
      }

    bid return_value (*this);
    if (return_value._denomination.is_notrump ())
      {
        return_value._denomination = clubs;
        return_value._level++;
      }
    else switch (return_value._denomination.get_suit ())
      {
      case clubs:
        return_value._denomination = diamonds;
        break;
      case diamonds:
        return_value._denomination = hearts;
        break;
      case hearts:
        return_value._denomination = spades;
        break;
      case spades:
        return_value._denomination.make_notrump ();
        break;
      default:
        assert (0);
      }

    return return_value;
  }

  bool operator==(const bid &other) const
  {
    return (this->_denomination == other._denomination
            &&
            this->_level        == other._level);
  }

  bool operator<(const bid &other) const
  {
    if (this->_level == other._level)
      return this->_denomination < other._denomination;
    else
      return this->_level < other._level;
  }
  friend std::ostream & operator<< (std::ostream &o, const bid &b);
};

// Capitalized so as not to conflict with the built-in
// floating-point type `double'
class Double: public call
{
  bool _fRedouble;
public:
  Double (bool redouble = false) : _fRedouble (redouble) {};
  unsigned short multiplier () const { return _fRedouble ? 4 : 2; }
  void print () const { std::cout << (_fRedouble ? "re" : "") << "double"; }
};

#endif  /* FILE_CALL_SEEN */
