// Copyright (C) 2001 Eric Hanchrow
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eric Hanchrow -- offby1@blarg.net
// 1914 N. 40th
// Seatlle, WA, USA
// 98103-8310

#if defined (WIN32)
# pragma warning(disable: 4786)
#endif

#include <vector>
#include <cstdlib>
#include <ctime>
#include <algorithm>
#include <fstream>
#include <string>

#if !defined (WIN32)
#include <argp.h>
#endif

#include <cerrno>
#include "card.hh"
#include "hand.hh"
#include "call.hh"
#include "auction.hh"
#include "dumb_player/dumb_player.hh"
#include "text_player/text_player.hh"
#include "compass.hh"
#include "trick.hh"
#include "score.hh"
#include "stats.hh"

// Number of times a card of the given rank has won a trick.  Aces,
// for example, typically win about a two-thirds of the tricks in
// which they're played, so the int value for rank ace will typically
// be 2/3 times the number of hands played.

// updated by trick::determine_winner.
std::map<rank, unsigned int> stats;

#if !defined (WIN32)
extern "C" {
const char * argp_program_version = PACKAGE " " VERSION;
const char * argp_program_bug_address = "offby1@blarg.net";

};
#endif

namespace {

void shuffle (std::vector<card> &v)
{
  for (unsigned short i=0; i<v.size (); i++)
    {
      int r = rand () % v.size ();
      card tmp = v[r];
      v[r] = v[i];
      v[i] = tmp;
    }
}

bool less_than (const card &c1, const card &c2)
{
  return ((c1.get_suit () == c2.get_suit ())
          ? (c1.get_rank () < c2.get_rank ())
          : (c1.get_suit () < c2.get_suit ()));
}

#if !defined (WIN32)
error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case ARGP_KEY_ARG:
      /* Too many arguments. */
      argp_usage (state);

      break;

    case ARGP_KEY_END:
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}
#endif

};

int
main (int argc, char *argv[])
{
#if !defined (WIN32)
  {
    std::string doc;

    doc += argp_program_version;
    doc += ", Copyright (C) 2001 Eric Hanchrow\n";

    doc += PACKAGE;
    doc += " comes with ABSOLUTELY NO WARRANTY.  ";
    doc += "This is free software,\nand you are welcome to redistribute it";
    doc += " under certain conditions; read the\nfile `COPYING'";
    doc += " in the source distribution, or at\n";
    doc += "http://www.gnu.org/licenses/gpl.html, for details.\n";

    doc += "\n";
    doc += PACKAGE;
    doc += " -- contract bridge over the Internet, or against your computer\v";
    doc += "This program lets you play a game of bridge with other players on the Internet, or against your computer.\n";

    static struct argp_option options[] = {
      { 0 }
    };

    struct argp our_argp = {options, parse_opt, 0, doc.c_str(), 0 };

    argp_parse (&our_argp, argc, argv, 0, 0, 0);
  }
#endif

  try
    {
      hand deck;

      // MS VC++ 6.0 apparently doesn't conform to the C++ standard:
      // it makes `i' visible outside of the `for' loop.  So I
      // contrain it with an extra set of curly braces.
      {
        for (unsigned short i=0; i<52; i++)
          {
            deck.push_back (card (i));
          }
      }

      // Good for production, but inconvenient for testing
      //srand (time (0));

      rubber_scoresheet scoresheet;

      // This doesn't include passed-out hands.  Thus the `continue'
      // statement (below) doesn't update it.
      unsigned int hands_played = 0;

      while (!scoresheet.rubber_is_over ())
        {
          dumb_player North;
          dumb_player East ;
          dumb_player South;
          text_player West ;
          std::vector<player *>Players;

          Players.push_back (&North);
          Players.push_back (&East );
          Players.push_back (&South);
          Players.push_back (&West );

          //shuffle (deck);
          North.pick_up_cards (hand (deck.begin () + 00, deck.begin () + 13));
          East .pick_up_cards (hand (deck.begin () + 26, deck.begin () + 39));
          South.pick_up_cards (hand (deck.begin () + 13, deck.begin () + 26));
          West .pick_up_cards (hand (deck.begin () + 39, deck.end   ()     ));

          auction the_auction;

          // Each player bids in turn.
          while (the_auction.get_state () == auction::in_progress)
            {
              player &current_player = *(Players[the_auction.size () % 4]);

              const seat &whoami = the_auction.dealer + the_auction.size ();
              const call *current_call = current_player.make_call (whoami,
                                                                   the_auction,
                                                                   scoresheet);
              the_auction.push_back (current_call);
            }

          switch (the_auction.get_state ())
            {
            case auction::passed_out:
              std::cout << "This hand was passed out." << std::endl;

              // It's OK that we're not updating hands_played, since
              // that variable isn't meant to record passed-out
              // hands.
              continue;
            case auction::contract_settled:
              std::cout << the_auction.get_contract () << std::endl;
              break;
            default:
              assert (0);
            }

          {
            const seat declarer = the_auction.get_contract ().get_declarer ();
            scoresheet.credit_honors (the_auction.get_contract (),

                                      // The declarer's two hands
                                      Players [(declarer + 0).offset_from_north ()]->get_hand(),
                                      Players [(declarer + 2).offset_from_north ()]->get_hand(),

                                      // The defender's two hands
                                      Players [(declarer + 1).offset_from_north ()]->get_hand (),
                                      Players [(declarer + 3).offset_from_north ()]->get_hand ());
          }

          plays tricks_played;

          // Since our `hands' array always has north as element 0, we
          // need to calculate the dummy's offset from north in order
          // to properly index the array.
          const hand &dummy = Players [(the_auction.get_contract ().get_declarer () + 2).offset_from_north ()]->get_hand ();

          // Play cards until north is out of cards.  Presumably the
          // other hands will run out at the same time :)
          while (North.get_hand().size ())
            {
              Assert<internal_error> (0 == tricks_played.size ()
                                      ||
                                      (tricks_played.back ().cards().size () == 4),
                                      "incomplete trick");

              const seat leader = ((tricks_played.size ()
                                    // most leads are made by player
                                    // who won the last trick
                                    ? tricks_played.back ().winner ()


                                    // Opening lead is made by player
                                    // to left of declarer
                                    : the_auction.get_contract ().get_declarer () + 1)

                                   );

              {
                std::vector<card> this_trick;

                // Each player plays a card to this trick.
                for (unsigned short cards_played = 0;
                     cards_played < 4;
                     cards_played++)
                  {
                    const seat &who = leader + cards_played;
                    player &current_player = *(Players[who.offset_from_north ()]);

                    card c = current_player.choose_card (the_auction,
                                                         scoresheet,
                                                         tricks_played,
                                                         this_trick,

                                                         // If this is
                                                         // the opening
                                                         // lead, we're
                                                         // not allowed
                                                         // to see the
                                                         // dummy's hand.
                                                         ((tricks_played.size ()
                                                           ||
                                                           cards_played)
                                                          ? &dummy
                                                          : 0));

                    this_trick.push_back (c);
                  }

                tricks_played.push_back (trick (leader,
                                                the_auction.get_contract ().get_bid ().denom (),
                                                this_trick));
              }
            }

          scoresheet.tally_hand (tricks_played,
                                 the_auction.get_contract ());

          if (scoresheet.rubber_is_over ())
            {
              std::cout << "Final score:"
                        << std::endl;
            }
          std::cout << "North-South now have "
                    << scoresheet.north_south ().above ()
                    << " points above the line, and "
                    << scoresheet.north_south ().below ()
                    << " below the line."
                    << std::endl
                    << "East-West have "
                    << scoresheet.east_west ().above ()
                    << " points above the line, and "
                    << scoresheet.east_west ().below ()
                    << " below the line."
                    << std::endl;
          hands_played++;
        }

      for (std::map<rank, unsigned int>::const_iterator i = stats.begin ();
           i != stats.end ();
           i++)
        {
          // 5 is enough to hold the longest ranks, namely `three',
          // `seven', `eight', and `queen'.
          std::streamsize old_width = std::cout.width (5);

          std::cout << i->first;

          std::cout.width (old_width); std::cout << "s won ";
          std::cout.width (5);         std::cout << i->second;
          std::cout.width (old_width); std::cout << " / ";
          std::cout.width (5);         std::cout << hands_played * 4;
          std::cout.width (old_width); std::cout <<" times ";
                                       std::cout << "(";
          std::cout.width (7);         std::cout << (100.0 * i->second) / hands_played / 4;
          std::cout.width (old_width); std::cout << "%)" << std::endl;
        }
    }
  catch (const std::string &s)
    {
      std::cerr << s << std::endl;
    }
  catch (const char *const s)
    {
      std::cerr << s << std::endl;
    }
  catch (bridge_exception &e)
    {
      std::cerr << "Someone tried to cheat! "
                << e.what()
                << std::endl;
    }
  catch (internal_error &e)
    {
      std::cerr << "Internal error: "
                << e.what()
                << std::endl;
    }
  catch (...)
    {
      std::cerr << "Hmph.  Unknown exception." << std::endl;
    }

  return EXIT_SUCCESS;
}
